from django.contrib.gis.shortcuts import compress_kml
from django.template import loader, Context
from rest_framework.renderers import BaseRenderer, JSONRenderer

from spillway import collections


class GeoJSONRenderer(JSONRenderer):
    """Renderer which serializes to GeoJSON.

    This renderer purposefully avoids reserialization of GeoJSON from the
    spatial backend which greatly improves performance.
    """
    media_type = 'application/vnd.geo+json'
    format = 'geojson'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """Returns *data* encoded as GeoJSON."""
        data = collections.as_feature(data)
        try:
            return data.geojson
        except AttributeError:
            return super(GeoJSONRenderer, self).render(
                data, accepted_media_type, renderer_context)


class TemplateRenderer(BaseRenderer):
    """Template based feature renderer."""
    template_name = None

    def render(self, data, accepted_media_type=None, renderer_context=None):
        collection = collections.as_feature(data)
        try:
            features = collection['features']
        except KeyError:
            features = [collection]
        template = loader.get_template(self.template_name)
        return template.render(Context({'features': features}))


class KMLRenderer(TemplateRenderer):
    """Renderer which serializes to KML."""
    media_type = 'application/vnd.google-earth.kml+xml'
    format = 'kml'
    template_name = 'spillway/placemarks.kml'


class KMZRenderer(KMLRenderer):
    """Renderer which serializes to KMZ."""
    media_type = 'application/vnd.google-earth.kmz'
    format = 'kmz'

    def render(self, *args, **kwargs):
        kmldata = super(KMZRenderer, self).render(*args, **kwargs)
        return compress_kml(kmldata)


class SVGRenderer(TemplateRenderer):
    """Renderer which serializes to SVG."""
    media_type = 'image/svg+xml'
    format = 'svg'
    template_name = 'spillway/features.svg'


class MapnikRenderer(BaseRenderer):
    """Renders Mapnik stylesheets to tiled PNG."""
    media_type = 'image/png'
    format = 'png'
    charset = None
    render_style = 'binary'

    def render(self, imgdata, accepted_media_type=None, renderer_context=None):
        # Just a no-op renderer as the mapnik map object should be serialized
        # prior to take full advantage of the caching framework.
        return imgdata


class MapnikJPEGRenderer(MapnikRenderer):
    """Renders Mapnik stylesheets to tiled JPEG."""
    media_type = 'image/jpeg'
    format = 'jpeg'



from rest_framework.response import Response
from rest_framework.generics import GenericAPIView, ListAPIView

from spillway import carto, filters, forms, mixins, renderers
from spillway.generics import BaseGeoView


class MapView(mixins.ResponseExceptionMixin, GenericAPIView):
    """View for rendering map tiles from /{z}/{x}/{y}/ tile coordinates."""
    renderer_classes = (renderers.MapnikRenderer,
                        renderers.MapnikJPEGRenderer)

    def get(self, request, *args, **kwargs):
        form = forms.RasterTileForm.from_request(request, view=self)
        m = carto.build_map([self.get_object()], form)
        # Mapnik Map object is not pickleable, so it breaks the caching
        # middleware. We must serialize the image before passing it off to the
        # Response and Renderer.
        return Response(m.render(request.accepted_renderer.format))


class TileView(BaseGeoView, ListAPIView):
    """View for serving tiled GeoJSON or PNG from a GeoModel."""
    pagination_class = None
    filter_backends = (filters.TileFilter,)
    renderer_classes = (renderers.GeoJSONRenderer, renderers.MapnikRenderer)

    def get(self, request, *args, **kwargs):
        if isinstance(request.accepted_renderer,
                      renderers.GeoJSONRenderer):
            return super(TileView, self).get(request, *args, **kwargs)
        form = forms.RasterTileForm.from_request(request, view=self)
        m = carto.build_map([self.get_queryset()], form)
        return Response(m.render(request.accepted_renderer.format))



from django import forms
from django.utils.deconstruct import deconstructible
from django.utils.translation import ugettext_lazy as _


@deconstructible
class GeometrySizeValidator(object):
    """Field validator to limit geometries to a given area."""
    message = _('Max area exceeded.')
    code = 'invalid'

    def __init__(self, max_area, srid=None):
        self.max_area = max_area
        self.srid = srid

    def __call__(self, geom):
        if self.srid and geom.srid != self.srid:
            geom.transform(self.srid)
        if (getattr(geom, 'ogr', geom).dimension > 1 and
                geom.area > self.max_area):
            raise forms.ValidationError(self.message, code=self.code)



from django.http import FileResponse
from django.forms import ValidationError as FormValidationError
from rest_framework import exceptions
from rest_framework.generics import ListAPIView, ListCreateAPIView, RetrieveAPIView
from rest_framework.serializers import ValidationError
from rest_framework.settings import api_settings
import rest_framework.renderers as rn

from spillway import filters, forms, mixins, pagination, renderers, serializers

_default_filters = tuple(api_settings.DEFAULT_FILTER_BACKENDS)
_default_renderers = tuple(api_settings.DEFAULT_RENDERER_CLASSES)


class BaseGeoView(mixins.ModelSerializerMixin):
    """Base view for models with geometry fields."""
    model_serializer_class = serializers.FeatureSerializer
    pagination_class = pagination.FeaturePagination
    filter_backends = _default_filters + (
        filters.SpatialLookupFilter, filters.GeoQuerySetFilter)
    renderer_classes = _default_renderers + (
        renderers.GeoJSONRenderer, renderers.KMLRenderer, renderers.KMZRenderer)


class GeoDetailView(BaseGeoView, RetrieveAPIView):
    """Generic detail view providing vector geometry representations."""


class GeoListView(BaseGeoView, ListAPIView):
    """Generic view for listing a geoqueryset."""


class GeoListCreateAPIView(BaseGeoView, ListCreateAPIView):
    """Generic view for listing or creating geomodel instances."""


class BaseRasterView(mixins.ModelSerializerMixin,
                     mixins.ResponseExceptionMixin):
    """Base view for raster models."""
    model_serializer_class = serializers.RasterModelSerializer
    filter_backends = _default_filters + (filters.RasterQuerySetFilter,)
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFZipRenderer,
        renderers.HFAZipRenderer,
    )

    def finalize_response(self, request, response, *args, **kwargs):
        response = super(BaseRasterView, self).finalize_response(
            request, response, *args, **kwargs)
        # Use streaming file responses for GDAL formats.
        if isinstance(getattr(response, 'accepted_renderer', None),
                      renderers.gdal.BaseGDALRenderer):
            headers = response._headers
            response = FileResponse(response.rendered_content)
            response._headers = headers
        return response

    def get_queryset(self):
        # Filter first so later RasterQuerySet methods always see a subset
        # instead of all available records.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg in self.kwargs:
            filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
            return self.queryset.filter(**filter_kwargs)
        return self.queryset.all()

    def options(self, request, *args, **kwargs):
        if isinstance(self.request.accepted_renderer,
                      renderers.gdal.BaseGDALRenderer):
            raise exceptions.NotAcceptable
        return super(BaseRasterView, self).options(request, *args, **kwargs)

    @property
    def paginator(self):
        # Disable pagination for GDAL Renderers.
        if not isinstance(self.request.accepted_renderer,
                          _default_renderers):
            self.pagination_class = None
        return super(BaseRasterView, self).paginator


class RasterDetailView(BaseRasterView, RetrieveAPIView):
    """View providing access to a Raster model instance."""
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFRenderer,
        renderers.HFARenderer,
    )


class RasterListView(BaseRasterView, ListAPIView):
    """View providing access to a Raster model QuerySet."""
    filter_backends = _default_filters + (
        filters.SpatialLookupFilter,
        filters.RasterQuerySetFilter,
    )



from django.core import exceptions
from django.contrib.gis import geos
from django.contrib.gis.db import models
from django.db.models.fields.files import FieldFile
from rest_framework import serializers
from greenwich.srs import SpatialReference

from spillway import query, collections as sc
from spillway.fields import GeometryField
from spillway.renderers.gdal import BaseGDALRenderer

serializers.ModelSerializer.serializer_field_mapping.update({
    models.GeometryField: GeometryField,
    models.PointField: GeometryField,
    models.LineStringField: GeometryField,
    models.PolygonField: GeometryField,
    models.MultiPointField: GeometryField,
    models.MultiLineStringField: GeometryField,
    models.MultiPolygonField: GeometryField,
    models.GeometryCollectionField: GeometryField
})


class GeoModelSerializer(serializers.ModelSerializer):
    """Serializer class for GeoModels."""

    def __new__(cls, *args, **kwargs):
        cls.Meta.geom_field = getattr(cls.Meta, 'geom_field', None)
        return super(GeoModelSerializer, cls).__new__(cls, *args, **kwargs)

    def get_fields(self):
        """Returns a fields dict for this serializer with a 'geometry' field
        added.
        """
        fields = super(GeoModelSerializer, self).get_fields()
        # Set the geometry field name when it's undeclared.
        if not self.Meta.geom_field:
            for name, field in fields.items():
                if isinstance(field, GeometryField):
                    self.Meta.geom_field = name
                    break
        return fields


class FeatureListSerializer(serializers.ListSerializer):
    """Feature list serializer for GeoModels."""

    @property
    def data(self):
        return super(serializers.ListSerializer, self).data

    def to_representation(self, data):
        data = map(self.child.to_representation, data)
        try:
            srid = query.get_srid(self.instance)
        except AttributeError:
            srid = None
        return sc.FeatureCollection(features=data, crs=srid)


class FeatureSerializer(GeoModelSerializer):
    """Feature serializer for GeoModels."""

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs['child'] = cls()
        meta = getattr(cls, 'Meta', None)
        list_serializer_cls = getattr(
            meta, 'list_serializer_cls', FeatureListSerializer)
        return list_serializer_cls(*args, **kwargs)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            self._data = super(FeatureSerializer, self).data
            if 'crs' not in self._data:
                try:
                    srid = getattr(self.instance, self.Meta.geom_field).srid
                except (AttributeError, geos.GEOSException):
                    pass
                else:
                    self._data['crs'] = sc.NamedCRS(srid)
        return self._data

    def to_representation(self, instance):
        native = super(FeatureSerializer, self).to_representation(instance)
        geometry = native.pop(self.Meta.geom_field)
        pk = native.pop(instance._meta.pk.name, None)
        return sc.Feature(pk, geometry, native)

    def to_internal_value(self, data):
        if sc.has_features(data):
            for feat in data['features']:
                return self.to_internal_value(feat)
        try:
            sref = SpatialReference(data['crs']['properties']['name'])
        except KeyError:
            sref = None
        # Force evaluation of fields property.
        if not self.fields and self.Meta.geom_field is None:
            raise exceptions.FieldDoesNotExist('Geometry field not found')
        record = {self.Meta.geom_field: data.get('geometry')}
        record.update(data.get('properties', {}))
        feature = super(FeatureSerializer, self).to_internal_value(record)
        if feature and sref:
            geom = feature[self.Meta.geom_field]
            geom.srid = sref.srid
        return feature


class RasterModelSerializer(GeoModelSerializer):
    """Serializer class for raster models."""

    def __new__(cls, *args, **kwargs):
        cls.Meta.raster_field = getattr(cls.Meta, 'raster_field', None)
        return super(RasterModelSerializer, cls).__new__(cls, *args, **kwargs)

    def get_fields(self):
        fields = super(RasterModelSerializer, self).get_fields()
        if not self.Meta.raster_field:
            for name, field in fields.items():
                if isinstance(field, serializers.FileField):
                    self.Meta.raster_field = name
                    break
        fieldname = self.Meta.raster_field
        request = self.context.get('request')
        renderer = getattr(request, 'accepted_renderer', None)
        try:
            obj = self.instance[0]
        except (IndexError, TypeError):
            obj = self.instance
        modelfield = getattr(obj, fieldname, None)
        if (isinstance(renderer, BaseGDALRenderer)
                or not isinstance(modelfield, FieldFile)):
            fields[fieldname] = serializers.ReadOnlyField()
        return fields



from django.http import HttpResponse


class MapnikResponse(HttpResponse):
    # media_types = {'png': 'image/png'}
    # def __init__(self, content, *args, **kwargs):
    # def __init__(self, mapobj, renderer, *args, **kwargs):
    def __init__(self, mapobj, renderer, **kwargs):
        kwargs.setdefault('content_type', renderer.media_type)
        super(MapnikResponse, self).__init__(
            content=mapobj.render(renderer.format), **kwargs)



import tempfile

from greenwich.io import MemFileIO


class SpooledCacheFile(tempfile.SpooledTemporaryFile):
    # cachedir = os.path.join('CACHE', 'images')
    cachedir = 'CACHE'

    # FIXME: Should duplicate constructor to avoid StringIO creation
    def __init__(self, dir=None, *args, **kwargs):
        dir = dir or self.cachedir
        super(SpooledCacheFile, self).__init__(dir=dir, *args, **kwargs)
        self._file = MemFileIO()


class SpooledFileIO(tempfile.SpooledTemporaryFile):
    """Temporary file wrapper switches from MemFileIO to an actual file when
    the maximum size is exceeded.
    """

    def __init__(self, max_size=10 * 1024, *args, **kwargs):
        tempfile.SpooledTemporaryFile.__init__(
            self, max_size=max_size, *args, **kwargs)
        self._file = MemFileIO()

    def _check(self, file):
        if self._rolled:
            return
        max_size = self._max_size
        # Seek to the end to be sure we really have a 0-length file.
        # if file.tell() == 0:
        if not file.tell():
            file.seek(0, 2)
        if max_size and file.tell() > max_size:
            self.rollover()

    def check_rollover(self):
    # def check_size(self):
        self._check(self._file)
        print 'ROLLED:', self._rolled, self.name

    def is_rolled(self):
        return self._rolled

    # FIXME: Drop this when MemFileIO has .getvalue() or not since we are
    # switching to NamedTemporaryFile.
    def rollover(self):
        if self._rolled:
            return
        file = self._file
        # newfile = self._file = tempfile.TemporaryFile(
        newfile = self._file = tempfile.NamedTemporaryFile(
            *self._TemporaryFileArgs)
        del self._TemporaryFileArgs
        # newfile.write(file.getvalue())
        # Seek first in absence of .getvalue()
        file.seek(0)
        newfile.write(file.read())
        # Why would the default implementation seek to the end? position is
        # already at the end after calling .write(), right?
        # newfile.seek(file.tell(), 0)
        # print 'NEWFILE:', newfile.tell()
        self._rolled = True



"""Serializer fields"""
from __future__ import absolute_import
import collections

from django.contrib.gis import geos, forms
from rest_framework import renderers
from rest_framework.fields import Field, FileField

from spillway.compat import json


class GeometryField(Field):
    def bind(self, field_name, parent):
        # Alter field source based on the requested format.
        try:
            renderer = parent.context['request'].accepted_renderer
        except (AttributeError, KeyError):
            pass
        else:
            obj = getattr(parent.context.get('view'), 'queryset',
                          parent.instance)
            if hasattr(obj, renderer.format):
                self.source = renderer.format
            elif isinstance(renderer, renderers.BrowsableAPIRenderer):
                self.source = '%s.wkt' % field_name
        super(GeometryField, self).bind(field_name, parent)

    def get_attribute(self, instance):
        # SpatiaLite returns empty/invalid geometries in WKT or GeoJSON with
        # exceedingly high simplification tolerances.
        try:
            return super(GeometryField, self).get_attribute(instance)
        except geos.GEOSException:
            return None

    def to_internal_value(self, data):
        # forms.GeometryField cannot handle geojson dicts.
        if isinstance(data, collections.Mapping):
            data = json.dumps(data)
        return forms.GeometryField().to_python(data)

    def to_representation(self, value):
        # Create a dict from the GEOSGeometry when the value is not previously
        # serialized from the spatial db.
        try:
            return {'type': value.geom_type, 'coordinates': value.coords}
        # Value is already serialized as geojson, kml, etc.
        except AttributeError:
            return value



from django.core.files.storage import default_storage
from django.db import connection
from django.contrib.gis import gdal
from greenwich import srs

from spillway.compat import mapnik
from spillway import colors, query

def make_dbsource(**kwargs):
    """Returns a mapnik PostGIS or SQLite Datasource."""
    if 'spatialite' in connection.settings_dict.get('ENGINE'):
        kwargs.setdefault('file', connection.settings_dict['NAME'])
        return mapnik.SQLite(wkb_format='spatialite', **kwargs)
    names = (('dbname', 'NAME'), ('user', 'USER'),
             ('password', 'PASSWORD'), ('host', 'HOST'), ('port', 'PORT'))
    for mopt, dopt in names:
        val = connection.settings_dict.get(dopt)
        if val:
            kwargs.setdefault(mopt, val)
    return mapnik.PostGIS(**kwargs)

def build_map(querysets, tileform):
    data = tileform.cleaned_data if tileform.is_valid() else {}
    stylename = data.get('style')
    m = Map()
    m.zoom_bbox(data.get('bbox'))
    for queryset in querysets:
        layer = m.layer(queryset, stylename)
        if isinstance(layer, RasterLayer):
            layer.add_colorizer_stops(data.get('limits'))
    return m


class Map(object):
    mapfile = default_storage.path('map.xml')

    def __init__(self):
        m = mapnik.Map(256, 256)
        try:
            mapnik.load_map(m, str(self.mapfile))
        except RuntimeError:
            pass
        m.buffer_size = 128
        m.srs = '+init=epsg:3857'
        self.map = m

    def layer(self, queryset, stylename=None):
        """Returns a map Layer.

        Arguments:
        queryset -- QuerySet for Layer
        Keyword args:
        stylename -- str name of style to apply
        """
        cls = VectorLayer if hasattr(queryset, 'geojson') else RasterLayer
        layer = cls(queryset, style=stylename)
        try:
            style = self.map.find_style(layer.stylename)
        except KeyError:
            self.map.append_style(layer.stylename, layer.style())
        layer.styles.append(layer.stylename)
        self.map.layers.append(layer._layer)
        return layer

    def render(self, format):
        img = mapnik.Image(self.map.width, self.map.height)
        mapnik.render(self.map, img)
        return img.tostring(format)

    def zoom_bbox(self, bbox):
        """Zoom map to geometry extent.

        Arguments:
        bbox -- OGRGeometry polygon to zoom map extent
        """
        try:
            bbox.transform(self.map.srs)
        except gdal.GDALException:
            pass
        else:
            self.map.zoom_to_box(mapnik.Box2d(*bbox.extent))


class Layer(object):
    """Base class for a Mapnik layer."""

    def __init__(self, queryset, style=None):
        table = str(queryset.model._meta.db_table)
        field = query.geo_field(queryset)
        sref = srs.SpatialReference(query.get_srid(queryset))
        layer = mapnik.Layer(table, sref.proj4)
        layer.datasource = make_dbsource(
            table=table, geometry_field=field.name)
        self._layer = layer
        self.stylename = style or self._layer.name
        self._symbolizer = None

    def __getattr__(self, attr):
        return getattr(self._layer, attr)

    def style(self):
        """Returns a default Style."""
        style = mapnik.Style()
        rule = mapnik.Rule()
        self._symbolizer = self.symbolizer()
        rule.symbols.append(self._symbolizer)
        style.rules.append(rule)
        return style

    def symbolizer(self):
        raise NotImplementedError


class RasterLayer(Layer):
    """A Mapnik layer for raster data types."""

    def __init__(self, obj, band=1, style=None):
        self._rstore = obj
        layer = mapnik.Layer(
            str(obj), srs.SpatialReference(obj.srs).proj4)
        layer.datasource = mapnik.Gdal(file=obj.image.path, band=band)
        self._layer = layer
        self.stylename = style or 'Spectral_r'
        self._symbolizer = None

    def add_colorizer_stops(self, limits):
        rcolors = colors.colormap.get(self.stylename)
        if rcolors:
            bins = self._rstore.linear(limits, k=len(rcolors))
            symbolizer = self._symbolizer
            for value, color in zip(bins, rcolors):
                symbolizer.colorizer.add_stop(value, mapnik.Color(color))

    def symbolizer(self):
        symbolizer = mapnik.RasterSymbolizer()
        symbolizer.colorizer = mapnik.RasterColorizer(
            mapnik.COLORIZER_LINEAR, mapnik.Color(0, 0, 0, 0))
        return symbolizer


class VectorLayer(Layer):
    """A Mapnik layer for vector data types."""

    def symbolizer(self):
        symbolizers = {
            mapnik.DataGeometryType.Point: mapnik.PointSymbolizer,
            mapnik.DataGeometryType.LineString: mapnik.LineSymbolizer,
            mapnik.DataGeometryType.Polygon: mapnik.PolygonSymbolizer
        }
        return symbolizers.get(self._layer.datasource.geometry_type(),
                               mapnik.PolygonSymbolizer)()



import os
import tempfile
import zipfile

from django.core import exceptions
from django.contrib.gis import geos
from django.contrib.gis.db.models import query
from django.db import connection, models
from django.utils.functional import cached_property
import greenwich
from greenwich.io import MemFileIO
import numpy as np

def filter_geometry(queryset, **filters):
    """Helper function for spatial lookups filters.

    Provide spatial lookup types as keywords without underscores instead of the
    usual "geometryfield__lookuptype" format.
    """
    fieldname = geo_field(queryset).name
    query = {'%s__%s' % (fieldname, k): v for k, v in filters.items()}
    return queryset.filter(**query)

def geo_field(queryset):
    # Try Django 1.8 syntax first, then fall back to 1.7 and below.
    try:
        return queryset._geo_field()
    except AttributeError:
        return queryset.query._geo_field()

def get_srid(queryset):
    """Returns the GeoQuerySet spatial reference identifier."""
    try:
        # Django 1.8
        srid = queryset.query.get_context('transformed_srid')
    except AttributeError:
        # Django<=1.7
        srid = queryset.query.transformed_srid
    return srid or queryset.geo_field.srid


# Many GeoQuerySet methods cannot be chained as expected and extending
# GeoQuerySet to work properly with serialization calls like
# .simplify(100).svg() will require patching
# django.contrib.gis.db.models.query. Work around this with a custom
# GeoQuerySet for now.
class GeoQuerySet(query.GeoQuerySet):
    """Extends the default GeoQuerySet with some unimplemented PostGIS
    functionality.
    """
    _scale = '%s(%%s, %%s, %%s)' % connection.ops.scale
    # Geometry outputs
    _formats = {'geojson': '%s(%%s, %%s)' % connection.ops.geojson,
                'kml': '%s(%%s, %%s)' % connection.ops.kml,
                'svg': '%s(%%s, 0, %%s)' % connection.ops.svg}

    def _as_format(self, sql, format=None, precision=8):
        val = self._formats.get(format, self._formats['geojson'])
        return self.extra(select={format: val % (sql, precision)})

    def _transform(self, srid=None):
        args, geo_field = self._spatial_setup('transform')
        if not srid:
            return args['geo_col']
        try:
            # Django 1.8
            self.query.add_context('transformed_srid', srid)
        except AttributeError:
            # Django<=1.7
            self.query.transformed_srid = srid
        return '%s(%s, %s)' % (args['function'], args['geo_col'], srid)

    def _trans_scale(self, colname, deltax, deltay, xfactor, yfactor):
        if connection.ops.spatialite:
            sql = 'ScaleCoords(ShiftCoords(%s, %.12f, %.12f), %.12f, %.12f)'
        else:
            sql = 'ST_TransScale(%s, %.12f, %.12f, %.12f, %.12f)'
        return sql % (colname, deltax, deltay, xfactor, yfactor)

    def _simplify(self, colname, tolerance=0.0):
        # connection.ops does not have simplify available for PostGIS.
        return ('ST_Simplify(%s, %s)' % (colname, tolerance)
                if tolerance else colname)

    def extent(self, srid=None):
        """Returns the GeoQuerySet extent as a 4-tuple.

        The method chaining approach of
        geoqset.objects.transform(srid).extent() returns the extent in the
        original coordinate system, this method allows for transformation.

        Keyword args:
        srid -- EPSG id for for transforming the output geometry.
        """
        if not srid and not connection.ops.spatialite:
            return super(GeoQuerySet, self).extent()
        transform = self._transform(srid)
        # Spatialite extent() is supported post-1.7.
        if connection.ops.spatialite:
            ext = {'extent': 'AsText(%s(%s))' % ('Extent', transform)}
        else:
            ext = {'extent': '%s(%s)' % (connection.ops.extent, transform)}
        # The bare order_by() is needed to remove the default sort field which
        # is not present in this aggregation. Empty querysets will return
        # [None] here.
        extent = (self.extra(select=ext)
                      .values_list('extent', flat=True)
                      .order_by()[0])
        if not extent:
            return ()
        try:
            try:
                # Django<=1.7
                return connection.ops.convert_extent(extent)
            except TypeError:
                # Django 1.8
                return connection.ops.convert_extent(extent, get_srid(self))
        except NotImplementedError:
            return geos.GEOSGeometry(extent, srid).extent

    def filter_geometry(self, **kwargs):
        """Convenience method for spatial lookup filters."""
        return filter_geometry(self, **kwargs)

    @property
    def geo_field(self):
        """Returns model geometry field."""
        return geo_field(self)

    def has_format(self, format):
        return format in self._formats

    def pbf(self, bbox, geo_col=None, scale=4096):
        """Returns tranlated and scaled geometries suitable for Mapbox vector
        tiles.
        """
        col = geo_col or self._transform()
        w, s, e, n = bbox.extent
        trans = self._trans_scale(col, -w, -s,
                                  scale / (e - w),
                                  scale / (n - s))
        return self.extra(select={'pbf': 'ST_AsText(%s)' % trans})

    def scale(self, x, y, z=0.0, tolerance=0.0, precision=8, srid=None,
              format=None, **kwargs):
        """Returns a GeoQuerySet with scaled and optionally reprojected and
        simplified geometries, serialized to a supported format.
        """
        if not any((tolerance, srid, format)):
            return super(GeoQuerySet, self).scale(x, y, z, **kwargs)
        transform = self._transform(srid)
        scale = self._scale % (transform, x, y)
        simplify = self._simplify(scale, tolerance)
        return self._as_format(simplify, format, precision)

    def simplify(self, tolerance=0.0, srid=None, format=None, precision=8):
        """Returns a GeoQuerySet with simplified geometries serialized to
        a supported geometry format.
        """
        # Transform first, then simplify.
        transform = self._transform(srid)
        simplify = self._simplify(transform, tolerance)
        if format:
            return self._as_format(simplify, format, precision)
        # TODO: EWKB bug is fixed in spatialite 4.2+ so this can be removed.
        if connection.ops.spatialite:
            # Spatialite returns additional precision when converting to wkt,
            # so avoid the call unless we are simplifying geometries.
            if not (tolerance or srid):
                return self
            simplify = 'AsEWKT(%s)' % simplify
        return self.extra(select={self.geo_field.name: simplify})

    def tile(self, bbox, tolerance=0.0, format=None, clip=False):
        """Returns a GeoQuerySet intersecting a tile boundary.

        Arguments:
        bbox -- tile extent as geometry
        Keyword args:
        tolerance -- geometry simplification tolerance
        format -- vector tile format as str
        clip -- clip geometries to tile boundary as boolean
        """
        clone = filter_geometry(self, intersects=bbox.ewkt)
        # Tile grid uses 3857, but coordinates should be in 4326 commonly.
        tile_srid = 3857
        coord_srid = bbox.srid
        transform = clone._transform()
        if clip:
            ewkt = ('GeomFromEWKT' if connection.ops.spatialite else
                    'ST_GeomFromEWKT')
            transform = "ST_Intersection(%s, %s('%s'))" % (
                transform, ewkt, bbox.ewkt)
        if clone.geo_field.srid != tile_srid:
            transform = 'ST_Transform(%s, %s)' % (transform, tile_srid)
        simplify = clone._simplify(transform, tolerance)
        latlng = 'ST_Transform(%s, %s)' % (simplify, coord_srid)
        if format == 'pbf':
            return clone.pbf(bbox, geo_col=latlng)
        return clone._as_format(latlng, format)


class RasterQuerySet(GeoQuerySet):
    def arrays(self, field_name=None):
        """Returns a list of ndarrays.

        Keyword args:
        field_name -- raster field name as str
        """
        fieldname = field_name or self.raster_field.name
        arrays = []
        for obj in self:
            arr = getattr(obj, fieldname)
            if isinstance(arr, np.ndarray):
                arrays.append(arr)
            else:
                arrays.append(obj.array())
        return arrays

    def aggregate_periods(self, periods):
        """Returns list of ndarrays averaged to a given number of periods.

        Arguments:
        periods -- desired number of periods as int
        """
        try:
            fieldname = self.raster_field.name
        except TypeError:
            raise exceptions.FieldDoesNotExist('Raster field not found')
        arrays = self.arrays(fieldname)
        arr = arrays[0]
        fill = getattr(arr, 'fill_value', None)
        if getattr(arr, 'ndim', 0) > 2:
            arrays = np.vstack(arrays)
        if len(arrays) > 1:
            marr = np.ma.array(arrays, fill_value=fill, copy=False)
        else:
            marr = arrays[0]
        # Try to reshape using equal sizes first and fall back to unequal
        # splits.
        try:
            means = marr.reshape((periods, -1)).mean(axis=1)
        except ValueError:
            means = [a.mean() for a in np.array_split(marr, periods)]
        obj = self[0]
        setattr(obj, fieldname, means)
        return [obj]

    def get(self, *args, **kwargs):
        # Need special handling of model instances with modified attributes,
        # otherwise they will be lost.
        if self._result_cache is not None:
            for obj in self._result_cache:
                for attr, val in kwargs.items():
                    if getattr(obj, attr) == val:
                        return obj
            raise self.model.DoesNotExist(
                '%s matching query does not exist.' %
                self.model._meta.object_name)
        return super(RasterQuerySet, self).get(*args, **kwargs)

    @cached_property
    def raster_field(self):
        """Returns the raster FileField instance on the model."""
        for field in self.model._meta.fields:
            if isinstance(field, models.FileField):
                return field
        return False

    def summarize(self, geom, stat=None):
        """Returns a new RasterQuerySet with subsetted/summarized ndarrays.

        Arguments:
        geom -- geometry for masking or spatial subsetting
        Keyword args:
        stat -- numpy supported summary stat as str (min/max/mean/etc)
        """
        if not hasattr(geom, 'num_coords'):
            raise TypeError('Need OGR or GEOS geometry, %s found' % type(geom))
        clone = self._clone()
        for obj in clone:
            arr = obj.array(geom)
            if arr is not None:
                if stat:
                    axis = None
                    if arr.ndim > 2:
                        axis = 1
                        arr = arr.reshape(arr.shape[0], -1)
                    arr = getattr(np.ma, stat)(arr, axis)
                if arr.size == 1:
                    arr = arr.item()
            obj.image = arr
        return clone

    def warp(self, format=None, srid=None, geom=None):
        """Returns a new RasterQuerySet with possibly warped/converted rasters.

        Keyword args:
        format -- raster file extension format as str
        geom -- geometry for masking or spatial subsetting
        srid -- spatial reference identifier as int for warping to
        """
        clone = self._clone()
        for obj in clone:
            obj.convert(format, geom)
            if srid:
                f = obj.image.file
                if isinstance(f, MemFileIO):
                    r = greenwich.Raster(f.name)
                else:
                    r = obj.raster()
                memio = MemFileIO(delete=False)
                dswarp = r.warp(srid, memio)
                obj.image.file = memio
                dswarp.close()
                r.close()
        return clone



import os

from django.http import FileResponse


class FileAttachmentResponse(FileResponse):
    def __init__(self, *args, **kwargs):
        super(FileAttachmentResponse, self).__init__(*args, **kwargs)
        fp = self.file_to_stream
        if hasattr(fp, 'name'):
            self.set_name(fp.name)
        length = kwargs.get('content_length')
        if length is None:
            # A little faster but won't work for filelike objects
            # os.fstat(f.fileno()).st_size
            pos = fp.tell()
            fp.seek(0, 2)
            length = fp.tell()
            fp.seek(pos)
        self['Content-Length'] = length

    # def set_length(self, length):

    def set_name(self, name):
        # self.setdefault(
            # 'Content-Disposition',
            # 'attachment; filename=%s' % os.path.basename(fp.name))
        fname = os.path.basename(name)
        self['Content-Disposition'] = 'attachment; filename=%s' % fname



from rest_framework import exceptions
from rest_framework.renderers import TemplateHTMLRenderer
from rest_framework.settings import api_settings


class ModelSerializerMixin(object):
    """Provides generic model serializer classes to views."""
    model_serializer_class = None

    def get_serializer_class(self):
        if self.serializer_class:
            return self.serializer_class
        class DefaultSerializer(self.model_serializer_class):
            class Meta:
                model = self.queryset.model
                fields = '__all__'
        return DefaultSerializer


class ResponseExceptionMixin(object):
    """Handle response exceptions by negotating among default renderers.

    The default exception handler passes error message dicts to the renderer
    which should be avoided with GDAL and Mapnik renderers, e.g. we wouldn't
    want a 404 returned as image/png.
    """

    def handle_exception(self, exc):
        response = super(ResponseExceptionMixin, self).handle_exception(exc)
        renderers = tuple(api_settings.DEFAULT_RENDERER_CLASSES)
        accepted = getattr(self.request, 'accepted_renderer', None)
        if (response.exception and accepted
                and not isinstance(accepted, renderers)):
            conneg = self.get_content_negotiator()
            try:
                render_cls, mtype = conneg.select_renderer(
                    self.request, renderers, self.format_kwarg)
            except exceptions.NotAcceptable:
                render_cls = TemplateHTMLRenderer
                mtype = render_cls.media_type
            self.request.accepted_renderer = render_cls()
            self.request.accepted_media_type = mtype
        return response



from .fields import (BoundingBoxField, CommaSepFloatField,
                     OGRGeometryField, SpatialReferenceField)
from .forms import (QuerySetForm, GeometryQueryForm, RasterTileForm,
                    VectorTileForm, RasterQueryForm, SpatialQueryForm)



import math

from django.contrib.gis import gdal, forms
from greenwich import tile
from rest_framework import renderers

from spillway import query
from spillway.compat import ALL_TERMS
from spillway.renderers import CSVRenderer
from . import fields


class QuerySetForm(forms.Form):
    """Base form for applying GeoQuerySet methods and filters."""

    def __init__(self, data=None, queryset=None, *args, **kwargs):
        super(QuerySetForm, self).__init__(data, *args, **kwargs)
        self.queryset = queryset
        self._is_selected = False

    @classmethod
    def from_request(cls, request, queryset=None, view=None):
        params = dict(request.query_params.dict(),
                      **getattr(view, 'kwargs', {}))
        params['format'] = request.accepted_renderer.format
        return cls(params, queryset)

    def query(self, force=False):
        """Returns the filtered/selected GeoQuerySet."""
        if not self.is_valid():
            raise forms.ValidationError(self.errors)
        if force:
            self._is_selected = False
        if not self._is_selected:
            if self.queryset is None:
                raise TypeError('Must be QuerySet not %s' %
                                type(self.queryset))
            self.select()
            self._is_selected = True
        return self.queryset

    def select(self):
        """Set GeoQuerySet from field values and filters.

        Subclasses may override this. Not called directly, use .query().
        """
        self.queryset = self.queryset.filter(**self.cleaned_data)


class SpatialQueryForm(QuerySetForm):
    """A Form for spatial lookup queries such as intersects, overlaps, etc.

    Includes 'bbox' as an alias for 'bboverlaps'.
    """
    bbox = fields.BoundingBoxField(required=False)

    def __init__(self, *args, **kwargs):
        super(SpatialQueryForm, self).__init__(*args, **kwargs)
        for lookup in self.data:
            if lookup in ALL_TERMS:
                self.fields[lookup] = fields.GeometryField(required=False)
                break

    def clean(self):
        cleaned_data = super(SpatialQueryForm, self).clean()
        spatial_lookup = set(cleaned_data.keys()) - {'bbox'}
        bbox = cleaned_data.pop('bbox', None)
        # Look for "bbox" which is just an alias to "bboverlaps".
        if bbox:
            cleaned_data.pop(spatial_lookup, None)
            cleaned_data['bboverlaps'] = bbox
        return cleaned_data

    def select(self):
        self.queryset = query.filter_geometry(
            self.queryset, **self.cleaned_data)


class GeometryQueryForm(QuerySetForm):
    """A form providing GeoQuerySet method arguments."""
    format = forms.CharField(required=False)
    precision = forms.IntegerField(required=False)
    # Tolerance value for geometry simplification
    simplify = forms.FloatField(required=False)
    srs = fields.SpatialReferenceField(required=False)

    def select(self):
        kwargs = {}
        data = self.cleaned_data
        if data['precision'] is not None:
            kwargs.update(precision=data['precision'])
        tolerance, srs, format = map(data.get, ('simplify', 'srs', 'format'))
        srid = getattr(srs, 'srid', None)
        try:
            has_format = self.queryset.has_format(format)
        except AttributeError:
            # Handle default GeoQuerySet.
            try:
                self.queryset = getattr(self.queryset, format)(**kwargs)
            except AttributeError:
                pass
        else:
            if has_format:
                kwargs.update(format=format)
            self.queryset = self.queryset.simplify(tolerance, srid, **kwargs)


class RasterQueryForm(QuerySetForm):
    """Validates format options for raster data."""
    bbox = fields.BoundingBoxField(required=False)
    format = forms.CharField(required=False)
    g = fields.OGRGeometryField(required=False)
    upload = fields.GeometryFileField(required=False)
    periods = forms.IntegerField(required=False)
    stat = forms.ChoiceField(
        choices=[(choice,) * 2 for choice in
                 'count', 'max', 'mean', 'median', 'min', 'std', 'var'],
        required=False)

    def clean(self):
        """Return cleaned fields as a dict, determine which geom takes
        precedence.
        """
        data = super(RasterQueryForm, self).clean()
        data['g'] = data.pop('upload') or data.pop('bbox') or data.get('g')
        return data

    def select(self):
        txtformats = (renderers.JSONRenderer.format, CSVRenderer.format)
        htmlformats = (renderers.BrowsableAPIRenderer.format,
                       renderers.TemplateHTMLRenderer.format)
        fields = ('format', 'g', 'stat', 'periods')
        format, geom, stat, periods = map(self.cleaned_data.get, fields)
        if not geom and format in htmlformats + txtformats:
            return
        elif geom and format in htmlformats:
            format = txtformats[0]
        if format in txtformats:
            qs = self.queryset.summarize(geom, stat)
        else:
            qs = self.queryset.warp(format, geom=geom)
        if periods:
            qs = qs.aggregate_periods(periods)
        self.queryset = qs


class TileForm(QuerySetForm):
    """Validates requested map tiling parameters."""
    bbox = fields.OGRGeometryField(srid=4326, required=False)
    size = forms.IntegerField(required=False, initial=256)
    x = forms.IntegerField()
    y = forms.IntegerField()
    z = forms.IntegerField()

    def clean(self):
        data = super(TileForm, self).clean()
        x, y, z = map(data.get, ('x', 'y', 'z'))
        # Create bbox from NW and SE tile corners.
        try:
            extent = (tile.to_lonlat(x, y, z) +
                      tile.to_lonlat(x + 1, y + 1, z))
        except ValueError:
            extent = (0, 0, 0, 0)
        geom = gdal.OGRGeometry.from_bbox(extent)
        geom.srid = self.fields['bbox'].srid
        data['bbox'] = geom
        return data


class RasterTileForm(TileForm):
    band = forms.IntegerField(required=False, initial=1)
    size = forms.IntegerField(required=False, initial=256)
    limits = fields.CommaSepFloatField(required=False)
    style = forms.CharField(required=False)

    def clean_band(self):
        return self.cleaned_data['band'] or self.fields['band'].initial

    def clean_style(self):
        # Mapnik requires string, not unicode, for style names.
        return str(self.cleaned_data['style'])


class VectorTileForm(TileForm):
    clip = forms.BooleanField(required=False, initial=False)
    format = forms.CharField(required=False)
    # Geometry simplification tolerances based on tile zlevel, see
    # http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames.
    tolerances = [6378137 * 2 * math.pi / (2 ** (zoom + 8))
                  for zoom in range(20)]

    def select(self):
        data = self.cleaned_data
        try:
            tolerance = self.tolerances[data['z']]
        except IndexError:
            tolerance = self.tolerances[-1]
        self.queryset = self.queryset.tile(
            data['bbox'], tolerance, data['format'], data['clip'])



# URL path separator, cannot rely on os.sep as it varies by platform.
sep = r'/'
# Regex for an optional file extension.
format_suffix = r'(?:\.(?P<format>[\w.]+))?'
# Regex groups for a tile URL in {z}/{x}/{y}.{format} order.
tile = ('(?P<z>\d+)',
        '(?P<x>\d+)',
        '(?P<y>\d+)%s/?$' % format_suffix)
tileregex = sep.join(tile)

def tilepath(regex):
    """Appends a tile path regex to a url path."""
    return r''.join((regex, tileregex))



from django.forms import ValidationError
from rest_framework import serializers
from rest_framework.filters import BaseFilterBackend

from spillway import forms


class FormFilterBackend(BaseFilterBackend):
    queryset_form = None

    def filter_queryset(self, request, queryset, view):
        form = self.queryset_form.from_request(request, queryset, view)
        try:
            return form.query()
        except ValidationError:
            raise serializers.ValidationError(form.errors)


class GeoQuerySetFilter(FormFilterBackend):
    """A Filter for calling GeoQuerySet methods."""
    queryset_form = forms.GeometryQueryForm


class RasterQuerySetFilter(FormFilterBackend):
    """A Filter for calling RasterQuerySet methods."""
    queryset_form = forms.RasterQueryForm


class SpatialLookupFilter(FormFilterBackend):
    """A Filter providing backend supported spatial lookups like intersects,
    overlaps, etc.
    """
    queryset_form = forms.SpatialQueryForm


class TileFilter(FormFilterBackend):
    queryset_form = forms.VectorTileForm



import sys
import itertools
import httplib
import socket
import urllib2
import urlparse

from django.core.urlresolvers import reverse, NoReverseMatch
from django.core.management.base import BaseCommand, CommandError
from django.test import Client
from django.utils.module_loading import import_string
from greenwich import tile, srs, Envelope


class Tiler(object):
# class TileSeeder(object):
# class TileRequest(s)(object):
    # def __init__(self, extent, zlevs, stderr=None):
    # def __init__(self, extent, zlevs, command=None):
    # def __init__(self, baseurl, tiles, command=None):
    def __init__(self, baseurl, tiles):
        # if len(baseurl) != 6:
            # raise ValueError('URL must be 6-length sequence')
        # self.baseurl = baseurl
        try:
            url = urlparse.urlunparse(baseurl)
            self.baseurl = baseurl
        except ValueError as exc:
            exc.args = ('URL must be 6-length sequence',)
            raise
        # self.tiles = tile_coords(extent, zlevs)
        self.tiles = tiles
        # FIXME: clean this up
        # self.stdout = sys.stdout
        # self.stderr = sys.stderr
        # self.cmd = command
        # self.urls = []
        self.format = 'png'
        self.tileview = 'raster-tiles'

    def _build_url(self, kwargs):
        try:
            urlpath = reverse(self.tileview, kwargs=kwargs)
        except NoReverseMatch as exc:
            raise CommandError(exc.message)
        # urlargs = ['', options['host'], urlpath, '', options['params'], '']
        # urlargs = (options['scheme'], options['host'], urlpath, '',
                   # options['params'], '')
        urlargs = list(self.baseurl)
        urlargs[2] = urlpath
        url = urlparse.urlunparse(urlargs)
        return url

    # def request_tiles(self, obj, options):
    def request_tiles(self, obj):
        for tile in self.tiles:
            # self.cmd.stdout.write('TILE z: %s, x: %s, y: %s' % tile)
            kwargs = dict(zip(('z', 'x', 'y'), tile),
                          slug=obj.slug, format=self.format)
            # url = self._build_url(obj.slug)
            url = self._build_url(kwargs)
            # self.cmd.stdout.write('Request: %s' % url)
            sys.stdout.write('Request: %s' % url)
            self._make_request(url)

    def _make_request(self, url):
        try:
            fp = urllib2.urlopen(url)
            print fp.headers.dict
        # TODO: Straighten out errors, more generic exc?
        # except urllib2.HTTPError as exc:
        except urllib2.URLError as exc:
            # self.cmd.stderr.write(self.cmd.style.ERROR(str(exc)))
            # self.cmd.stderr.write(self.cmd.style.ERROR('%s: "%s"' % (exc, url)))
            sys.stderr.write(self.cmd.style.ERROR('%s: "%s"' % (exc, url)))
        except httplib.HTTPException as exc:
            # self.cmd.stderr.write(self.cmd.style.ERROR('XFailed: "%s"' % exc))
            sys.stderr.write(self.cmd.style.ERROR('XFailed: "%s"' % exc))
        except socket.timeout as exc:
            # self.cmd.stderr.write(self.cmd.style.ERROR('Failed: "%s"' % exc))
            sys.stderr.write(self.cmd.style.ERROR('Failed: "%s"' % exc))
        # except (httplib.HTTPException, socket.timeout, urllib2.URLError) as exc:
            # self.cmd.stderr.write(self.cmd.style.ERROR(str(exc)))
        else:
            fp.close()

# tilestache-seed.py -c ./config.json -l osm -b 37.79 -122.35 37.83 -122.25 -e png 12 13 14 15
class Command(BaseCommand):
    help = 'Preseed a tile cache'

    # TODO: Need (opt)args for ModelName, view-name, slug/layer, bbox, format
    def add_arguments(self, parser):
        parser.add_argument('model')
        # parser.add_argument('slug')
        parser.add_argument('zoom_levels', nargs='+', type=int)
        parser.add_argument(
            # '-L', '--layer-slug',
            '-s', '--slug',
            # help='Layer slug',
        )
        # parser.add_argument(
            # '--model',
            # help='Model',
        # )
        parser.add_argument('-H', '--host', default='localhost')
        parser.add_argument('-S', '--scheme', default='http')
        parser.add_argument('-p', '--params')
        # parser.add_argument('-q', '--query')
        # concurrent requests
        # parser.add_argument('-j', '--jobs')

    def handle(self, *args, **options):
        try:
            model = import_string(options['model'])
        except ImportError:
            raise CommandError('Cannot import Model: "%s"' % options['model'])
        qs = model.objects.all()
        print 'ARGS:', args
        print 'OPTIONS:', options
        if options['slug']:
            qs = qs.filter(slug=options['slug'])
        if not qs.count():
            # raise CommandError('No results for slug: "%s"' % options['slug'])
            raise CommandError('Empty QuerySet for "%s"' % model)
        # slug = 'slr_sfbay_0_0m_jdr'
        # qs = qs.filter(slug=slug)
        for obj in qs:
            # list_tiles()
            tiles = tile.from_bbox(obj.geom.extent, options['zoom_levels'])
            # self.request_tiles(obj, tiles, options)
            urlargs = (options['scheme'], options['host'], '', '',
                       options['params'], '')
            t = Tiler(urlargs, tiles, command=self)
            # t = Tiler(urlargs, obj.geom.extent,
                      # options['zoom_levels'], command=self)
            t.request_tiles(obj)



from .renderers import (GeoJSONRenderer, TemplateRenderer, KMLRenderer,
                        KMZRenderer, SVGRenderer, MapnikRenderer,
                        MapnikJPEGRenderer)
from .gdal import (CSVRenderer, GeoTIFFRenderer, GeoTIFFZipRenderer,
                   HFARenderer, HFAZipRenderer, JPEGRenderer,
                   JPEGZipRenderer, PNGRenderer, PNGZipRenderer)
